;-----------------------------------------------------------------------------
;
; ph_.asm ver 1.4 (c) PaulHoule.com  10/25/2019
;
; Selected ph_.c functions in assembly.
; Compiled w/JWASM2.12 -5 -coff (32-bit) or -win64 (64-bit)
;
; Objective is code size, not performance.  No Win64 SEH info is generated.
;
; These functions are simple enough that macros are used to have one source
; for both 32 and 64 bit.  This approach discourages using r8-r15, which
; yields a codespace bonus, as opcodes (eg push/pop, or those with dword
; operand size) using r8-r15 are larger.
;
; A 64-bit function will call another internal function without reserving
; shadow space; none of our .asm functions require shadow space.
;
;-----------------------------------------------------------------------------

if36	macro	a32, a64		;return a32 or a64 if 32/64 bit
	ifdef	rax			;rax defined only if 64-bit
	  exitm	<a64>
	endif
	exitm	<a32>
	endm

xax	textequ	if36(eax, rax)
xbx	textequ	if36(ebx, rbx)
xcx	textequ	if36(ecx, rcx)
xdx	textequ	if36(edx, rdx)
xsi	textequ	if36(esi, rsi)
xdi	textequ	if36(edi, rdi)
xbp	textequ	if36(ebp, rbp)
xsp	textequ	if36(esp, rsp)
size_t	textequ if36(dword, qword)

;---------------------------------------------------------------------
;==== need an implementation of __chkstk in here, or document
;====   user needs to add compiler chkstk.obj file to link list?
;---------------------------------------------------------------------

	;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; Couldn't get an assembler (eg JWASM 2.12, UASM 2.49) to build what
	; we wanted via proc, so proc is essentially shut off -- arg loading
	; and register push/pop must be done manually.  (We do require
	; the proc to decorate the name and make it public.)
	;
	; Another thing required out of proc is that, in 32-bit mode, ret
	; generates a "retn amount" with the correct 'amount' value (which
	; is determined by argument count).  The 64-bit retn 'amount'
	; is always zero (caller removes stacked args/shadow space).
	;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	if36(<.model flat,fastcall>)
	.listall
	option	stackbase:xsp
	if36(,<option prologue:none>)
	if36(,<option epilogue:none>)

proc_	macro	name:label, args:vararg	;used to build COMDAT enclosed funcs
	.errnz	@InStr(0, <args>,<uses>), <name: uses not allowed>
_TEXT_&name&	segment	byte flat alias(".text") comdat(1) 'CODE'
name		proc	args
	endm
endp_	macro	name:label
name		endp
_TEXT_&name&	ends
	endm

abyte	macro	aby, awo, adw		;abyte/aword/adword/awchr: used
	exitm	<aby>			;  by routine building macros to
	endm				;  generate code based on argument
aword	macro	aby, awo, adw		;  size (byte or wide-char)
	exitm	<awo>
	endm
adword	macro	aby, awo, adw
	exitm	<adw>
	endm
awchr	textequ	<aword>			;aword/adword per type of wide char


issp	macro	name:label, abwd
name	proc_	chr:abwd(byte, word, dword)
	abwd(<mov eax,ecx>,<movzx eax,cx>,<mov eax,ecx>)
	.if	abwd(al, eax, eax) != ' '
	  sub	abwd(al, eax, eax),9
	  sub	abwd(al, eax, eax),5
	  sbb	eax,eax
	.endif
	ret
name	endp_
	endm
isspace	 issp	abyte
iswspace issp	awchr

	; The asm "rep" versions of string funcs are faster -- 60% of
	; the time of loops generated by MSVC 8 (timed on i5-2500k).

mcmp	macro	name:label, abwd
name	proc_	pDst:ptr, pSrc:ptr, n:size_t
	push	xsi
	mov	xsi,xcx			;xsi= pDst
	push	xdi
	mov	xdi,xdx			;xdi= pSrc
	xor	eax,eax
	mov	xcx,if36([esp+3*4], r8)	;xcx= n
	repe abwd(cmpsb, cmpsw, cmpsd)
	.if	!zero?
	  sbb	xax,xax
	  jnz	@f
	    inc	eax
	  @@:
	.endif
	pop	xdi
	pop	xsi
	ret
name	endp_
	endm
memcmp	mcmp	abyte
wmemcmp	mcmp	awchr

if 0 ;?memcpy() unused -- ?memcpy #define'd to ?memmove in ph_.h
mcpy	macro	name:label, abwd
name	proc_	pDst:ptr, pSrc:ptr, n:size_t
	push	xdi
	mov	xdi,xcx			;xdi= pDst
	push	xsi
	mov	xsi,xdx			;xsi= pSrc
	mov	xax,xcx
	mov	xcx,if36([esp+3*4], r8)	;xcx= n
	rep abwd(movsb, movsw, movsd)
	pop	xsi
	pop	xdi
	ret
name	endp_
	endm
memcpy	mcpy	abyte
wmemcpy	mcpy	awchr
endif

;Note: moving backward (low->hi addr) is slower, so we try hard not to
mmov	macro	name:label, abwd
name	proc_	pDst:ptr, pSrc:ptr, n:size_t
	push	xdi
	mov	xdi,xcx			;xdi= pDst
	push	xsi
	mov	xsi,xdx			;xsi= pSrc
	mov	xax,xcx
	mov	xcx,if36([esp+3*4], r8)	;xcx= n
	sub	xdx,xdi			;carry if backward (lo->hi move)
	.if	carry?			;if may need to reverse direction,
	  add	xdx,xcx			;xdx= if > 0, size of region overlap
	  ja	@f			;if overlap reverse direction,
	    lea	xdi,[xdi+abwd(xcx-1, xcx*2-2, xcx*4-4)]
	    lea	xsi,[xsi+abwd(xcx-1, xcx*2-2, xcx*4-4)]
	    std
	  @@:
	.endif
	rep abwd(movsb, movsw, movsd)
	cld
	pop	xsi
	pop	xdi
	ret
name	endp_
	endm
memmove	 mmov	abyte
wmemmove mmov	awchr

mset	macro	name:label, abwd
name	proc_	pDst:ptr, chr:abwd(byte, word, dword), n:size_t
	push	xdi
	mov	xdi,xcx			;xdi= pDst
	push	xcx
	mov	eax,edx			;xax= chr
	mov	xcx,if36([esp+3*4], r8)	;xcx= n
	rep abwd(stosb, stosw, stosd)
	pop	xax
	pop	xdi
	ret
name	endp_
	endm
memset	mset	abyte
wmemset	mset	awchr

mchr	macro	name:label, abwd
name	proc_	pDst:ptr, chr:abwd(byte, word, dword), n:size_t
	push	xdi
	mov	xdi,xcx			;xdi= pDst
	mov	eax,edx			;xax= chr
	mov	xcx,if36([esp+2*4], r8)	;xcx= n
	if36 (jecxz, jrcxz) @f
	repne abwd(scasb, scasw, scasd)
	lea	xax,[xdi-abwd(1, 2, 4)]
	.if	!zero?
	  @@: xor eax,eax
	.endif
	pop	xdi
	ret
name	endp_
	endm
memchr	mchr	abyte
wmemchr	mchr	awchr

slen	macro	name:label, abwd
name	proc_	pSrc:ptr
	push	xdi
	mov	xdi,xcx			;xdi= pSrc
	xor	eax,eax
	or	xcx,-1
	repne abwd(scasb, scasw, scasd)
	sub	xax,xcx
	pop	xdi
	if36(<dec eax>, <sub xax,2>)
	if36(<dec eax>)
	ret
name	endp_
	endm
strlen	slen	abyte
wcslen	slen	awchr

scpy	macro	name:label, abwd
name	proc_	pDst:ptr, pSrc:ptr
	push	xcx
	sub	xcx,xdx
	.repeat
	  abwd(<mov al,>, <movzx eax,word ptr>, <mov eax,>) [xdx]
	  mov	[xcx+xdx],abwd(al, ax, eax)
	  abwd(<inc xdx>, <add xdx,2>, <add xdx,4>)
	.until	!abwd(al, eax, eax)
	pop	xax
	ret
name	endp_
	endm
strcpy	scpy	abyte
wcscpy	scpy	awchr

sncpy macro	name:label, abwd
name	proc_	pDst:ptr, pSrc:ptr, n:size_t
	push	xdi
	mov	xdi,xcx			;xdi= pDst
	push	xcx
	mov	xcx,if36([esp+3*4], r8)	;xcx= n
	sub	xdx,xdi			;xdx= pSrc - pDst
	if36 (jecxz, jrcxz) @f
	  .repeat
	    abwd(<mov al,>, <movzx eax,word ptr>, <mov eax,>) [xdx+xdi]
	    .break .if !abwd(al, eax, eax)
	    abwd(stosb, stosw, stosd)
	    dec	xcx
	  .until zero?
	  rep abwd(stosb, stosw, stosd)
	@@:
	pop	xax
	pop	xdi
	ret
name	endp_
	endm
strncpy	sncpy abyte
wcsncpy	sncpy awchr

scmp	macro	name:label, abwd
name	proc_	pDst:ptr, pSrc:ptr
	push	xbx
	sub	xdx,xcx
	.repeat
	  abwd(movzx, movzx, mov) eax,abwd(byte, word, dword) ptr [xcx]
	  abwd(movzx, movzx, mov) ebx,abwd(byte, word, dword) ptr [xdx+xcx]
	  abwd(<inc xcx>, <add xcx,2>, <add xcx,4>)
	  sub	eax,ebx
	  .break .if !zero?
	.until	!ebx
	if36(,<movsxd rax,eax>)
	pop	xbx
	ret
name	endp_
	endm
strcmp	scmp	abyte
wcscmp	scmp	awchr

scat	macro	name:label, abwd
name	proc_	pDst:ptr, pSrc:ptr
	push	xcx
	xor	eax,eax
	push	xdi
	mov	xdi,xcx
	.if	[xdx] != abwd(al, ax, eax)
	  or	xcx,-1
	  repne abwd(scasb, scasw, scasd)
	  abwd(<dec xdi>, <sub xdi,2>, <sub xdi,4>)
	  sub	xdi,xdx
	  .repeat
	    abwd(<mov cl,>, <movzx ecx,word ptr>, <mov ecx,>) [xdx]
	    mov	[xdi+xdx],abwd(cl, cx, ecx)
	    abwd(<inc xdx>, <add xdx,2>, <add xdx,4>)
	  .until !abwd(cl, ecx, ecx)
	.endif
	pop	xdi
	pop	xax
	ret
name	endp_
	endm
strcat	scat	abyte
wcscat	scat	awchr

snct	macro	name:label, abwd
name	proc_	pDst:ptr, pSrc:ptr, n:size_t
	push	xcx
	xor	eax,eax
	if36(<push ebx>)		;ebx/r8= n
	if36(<mov ebx,[esp+3*4]>)
	push	xdi
	mov	xdi,xcx			;edi= pDst,  edx= pSrc
	.if	if36(ebx, r8) && [xdx] != abwd(al, ax, eax)
	  or	xcx,-1
	  repne abwd(scasb, scasw, scasd)
	  abwd(<dec xdi>, <sub xdi,2>, <sub xdi,4>)
	  sub	xdx,xdi
	  .repeat
	    abwd(<mov cl,>, <movzx ecx,word ptr>, <mov ecx,>) [xdx+xdi]
	    .break .if !abwd(cl, ecx, ecx)
	    mov	[xdi],abwd(cl, cx, ecx)
	    abwd(<inc xdi>, <add xdi,2>, <add xdi,4>)
	    dec	if36(ebx, r8)
	  .until zero?
	  abwd(stosb, stosw, stosd)
	.endif
	pop	xdi
	if36(<pop ebx>)
	pop	xax
	ret
name	endp_
	endm
strncat	snct	abyte
wcsncat	snct	awchr

schr	macro	name:label, abwd
name	proc_	pSrc:ptr, chr:abwd(byte, word, dword)
	.repeat				;xd?= chr
	  abwd(<mov al,>, <movzx eax,word ptr>, <mov eax,>) [xcx]
	  abwd(<inc xcx>, <add xcx,2>, <add xcx,4>)
	  cmp	abwd(<al,dl>, <eax,edx>, <eax,edx>)
	  jz	@f
	.until	!abwd(al, eax, eax)
	xor	 eax,eax
	ret
@@:	lea	xax,[xcx - abwd(1, 2, 4)]
	ret
name	endp_
	endm
strchr	schr	abyte
wcschr	schr	awchr

	end
